import random
from importlib import resources

import pygame


class Map:
    def __init__(self, map_size):
        self.tile_size = 64
        self.map_width = map_size[0]
        self.map_height = map_size[1]
        self.tiles = []

        with resources.path("flyover.assets.tiles", "Grass.png") as path:
            self.tiles.append(path)
        with resources.path("flyover.assets.tiles", "Sea.png") as path:
            self.tiles.append(path)

        self.tile_map = []
        for column in range(0, self.map_width, self.tile_size):
            for row in range(0, self.map_height, self.tile_size):
                tile_coordinate = (column, row)
                tile = self.select_tile()
                self.tile_map.append((tile_coordinate, pygame.image.load(tile)))

    def select_tile(self):
        # TODO draw more terrain tiles and come up with code to arrange semi - not totally randomly
        tile = random.choice(self.tiles)
        return tile



import math
import random
from importlib import resources

import pygame

class Camera:
    def __init__(self, player, window_size, map):
        self.camera_width = window_size[0]
        self.camera_height = window_size[1]
        self.player = player
        self.map = map
        self.move()

    def move(self):
        self.rect = (self.player.collision_box.centerx - int(self.camera_width / 2),
                     self.player.collision_box.centery - int(self.camera_height / 2),
                     self.player.collision_box.centerx + int(self.camera_width / 2),
                     self.player.collision_box.centery + int(self.camera_height / 2))

    def active_tiles(self):
        # returns a list of tile tuples(x,y,img) that are currently around the player
        # and within the range of the camera
        active_tiles = []
        for tile in self.map.tile_map:
            if tile[0][0] in range(self.rect[0] - 64, self.rect[2] + 64):
                if tile[0][1] in range(self.rect[1] - 64, self.rect[3] + 64):
                    active_tiles.append(tile)
        return active_tiles


class Jet:
    speed = 2
    acceleration = 1
    rotation_speed = 5
    maximum_speed = 20
    minimum_speed = 2
    sprite_size = (64, 64)

    def __init__(self, starting_coordinates, map_size):
        self.angle = starting_coordinates[2]
        # x and y = center of entity, blitx and blity = top left corner
        # angle defined as 0 degrees North, 90 deg East etc.
        # Collision box defined as the inner 1/4 of the sprite
        self.collision_box = pygame.Rect((starting_coordinates[0] - self.sprite_size[0] / 4,
                                          starting_coordinates[1] - self.sprite_size[1] / 4),
                                         (self.sprite_size[0] / 2, self.sprite_size[1] / 2))
        # initialise game and blit coordinates
        self.update_coordinates()
        self.map_size = map_size
        self.state = 'level'

    def update_coordinates(self):
        # Draw x and draw y give the top left of the sprite - the coordinates passed to blit
        self.blit_x = self.collision_box.centerx - self.sprite_size[0] / 2
        self.blit_y = self.collision_box.centery - self.sprite_size[1] / 2
        self.x = self.collision_box.centerx
        self.y = self.collision_box.centery
        self.coord = (self.x, self.y)

    def stay_within_map(self):
        if self.x <= 0:
            self.angle = 90
        if self.y <= 0:
            self.angle = 180
        if self.x >= self.map_size[0]:
            self.angle = 270
        if self.y >= self.map_size[1]:
            self.angle = 0

    def within_active_area(self, camera):
        if self.x in range(camera.rect[0], camera.rect[2]):
            if self.y in range(camera.rect[1], camera.rect[3]):
                return True

    def move(self):
        x_diff = int(self.speed * math.sin(math.radians(self.angle)))
        y_diff = int(self.speed * math.cos(math.radians(self.angle))) * -1
        self.collision_box.move_ip(x_diff, y_diff)
        self.update_coordinates()
        self.stay_within_map()

    def normalize_angle(self, angle):
        if angle < 0:
            angle += 360
        elif angle >= 360:
            angle -= 360
        return int(angle)

    def rotate(self, direction):
        self.angle += direction * self.rotation_speed
        self.angle = self.normalize_angle(self.angle)

    def accelerate(self, direction):
        if self.speed == self.maximum_speed:
            if direction == -1:
                self.speed += direction * self.acceleration
        elif self.speed == self.minimum_speed:
            if direction == 1:
                self.speed += direction * self.acceleration
        else:
            self.speed += direction * self.acceleration

    def player_angle(self, player):
        # tells you what angle entity needs to point at to face player
        dist, x_diff, y_diff = self.distance_to_player(player)
        if dist == 0:  # Currently at the player's position
            return int(player.angle)

        angle = math.degrees(math.acos(x_diff / dist))
        if angle == 0:
            return 90
        elif angle == 180:
            return 270
        if y_diff > 0:
            angle += 90
        elif y_diff < 0:
            angle = 90 - angle

        return int(self.normalize_angle(angle))

    def distance_to_player(self, player):
        x_diff = player.x - self.x
        y_diff = player.y - self.y
        dist = math.hypot(x_diff, y_diff)
        return dist, x_diff, y_diff

    def shoot_missile(self):
        return Missile((self.x, self.y, self.angle), str(self), self.map_size)


class Player(Jet):
    def __init__(self, starting_coordinates, map_size):
        Jet.__init__(self, starting_coordinates, map_size)
        with resources.path("flyover.assets.sprites", "player.png") as path:
            self.sprite_level = pygame.image.load(path)
        with resources.path("flyover.assets.sprites", "player_left.png") as path:
            self.sprite_left = pygame.image.load(path)
        with resources.path("flyover.assets.sprites", "player_right.png") as path:
            self.sprite_right = pygame.image.load(path)

    def __str__(self):
        return 'Player'


class Enemy(Jet):
    def __init__(self, starting_coordinates, map_size):
        Jet.__init__(self, starting_coordinates, map_size)
        self.behaviours = (self.turn_left, self.turn_right, self.do_nothing,
                           self.speed_up, self.slow_down, self.follow_player)
        self.speed = 2
        self.firing_range = 200
        # starting behaviours
        self.behaviour = self.do_nothing
        self.last_behaviour = self.do_nothing
        self.last_behaviour_time = 0
        self.hit = False

    def __repr__(self):
        return 'Enemy @ ' + str(self.coord)

    def __str__(self):
        return 'Enemy'

    def move(self):
        super().move()
        # add to the behaviour counter after every move
        self.last_behaviour_time += 1

    def choose_behaviour(self, player):  # All behaviours need optional kwargs so player can be passed to follow player
        if self.last_behaviour == self.turn_left:
            self.behaviour = self.do_nothing
        elif self.last_behaviour == self.turn_right:
            self.behaviour = self.do_nothing
        else:
            self.behaviour = random.choice(self.behaviours)
            self.behaviour_duration = random.randint(5, 30)
        self.behaviour(player=player)
        self.last_behaviour = self.behaviour

    def turn_left(self, **kwargs):
        if self.last_behaviour_time < self.behaviour_duration:
            self.rotate(-1)
            self.state = 'left'
        else:
            self.state = 'level'

    def turn_right(self, **kwargs):
        if self.last_behaviour_time < self.behaviour_duration:
            self.rotate(1)
            self.state = 'right'
        else:
            self.state = 'level'

    def slow_down(self, **kwargs):
        self.state = 'level'
        if self.last_behaviour_time < int(self.behaviour_duration / 10):  # /10 to avoid hitting min/max speed in one go
            self.accelerate(-1)

    def speed_up(self, **kwargs):
        self.state = 'level'
        if self.last_behaviour_time < int(self.behaviour_duration / 10):
            self.accelerate(1)

    def follow_player(self, player):
        if self.last_behaviour_time < int(self.behaviour_duration):
            target_angle = self.player_angle(player)
            angle_change = abs(target_angle - self.angle)
            # check to make sure won't try to turn more than 180 degrees
            opposite = False  # if angle is more than 180 it's quicker to go the other way
            if angle_change > 180:
                opposite = True
            elif angle_change == 180:  # don't get stuck flying away from player
                self.turn_right()
            # choose direction to turn
            if self.angle > target_angle:
                if opposite:
                    self.turn_right()
                else:
                    self.turn_left()
            elif self.angle < target_angle:
                if opposite:
                    self.turn_left()
                else:
                    self.turn_right()

    def do_nothing(self, **kwargs):
        self.state = 'level'
        pass

    def check_sights(self, player):
        dist = self.distance_to_player(player)[0]
        if dist <= self.firing_range:
            # check nose is +- 10 degrees of player
            if self.angle in range(self.player_angle(player) - 10, self.player_angle(player) + 10):
                return self.shoot_missile()


class Missile(Jet):
    speed = Jet.maximum_speed + 10
    fuse = 20

    def __init__(self, starting_coordinates, shooter, map_size):
        Jet.__init__(self, starting_coordinates, map_size)
        self.x = starting_coordinates[0]  # start in sprite's center
        self.y = starting_coordinates[1]
        self.angle = starting_coordinates[2]
        self.shooter = shooter
        self.time_alive = 0
        width, height = 2, 6
        if self.angle in range(315, 360):
            x, y = width, height
        elif self.angle in range(0, 46):
            x, y = width, height
        elif self.angle in range(135, 226):
            x, y = width, height
        else:
            x, y = height, width
        self.surface = pygame.Surface((x, y))
        self.surface.fill((255, 255, 255))
        self.collision_box = pygame.Rect((self.x - 1, self.y - 1), (width, height))

    def __str__(self):
        return 'Missile shot by ' + self.shooter

    def move(self):
        super().move()
        self.time_alive += 1

    def check_hits(self, entity):
        if self.collision_box.colliderect(entity.collision_box):
            return True

    def stay_within_map(self):
        pass


class Mig35(Enemy):
    def __init__(self, starting_coordinates, map_size):
        Enemy.__init__(self, starting_coordinates, map_size)
        with resources.path("flyover.assets.sprites", "Mig35.png") as path:
            self.sprite_level = pygame.image.load(path)
        with resources.path("flyover.assets.sprites", "Mig35_left.png") as path:
            self.sprite_left = pygame.image.load(path)
        with resources.path("flyover.assets.sprites", "Mig35_right.png") as path:
            self.sprite_right = pygame.image.load(path)



class Mig21(Enemy):
    def __init__(self, starting_coordinates, map_size):
        Enemy.__init__(self, starting_coordinates, map_size)
        with resources.path("flyover.assets.sprites", "Mig21.png") as path:
            self.sprite_level = pygame.image.load(path)
        with resources.path("flyover.assets.sprites", "Mig21_left.png") as path:
            self.sprite_left = pygame.image.load(path)
        with resources.path("flyover.assets.sprites", "Mig21_right.png") as path:
            self.sprite_right = pygame.image.load(path)



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "readme.md").read_text()

# This call to setup() does all the work
setup(
    name="flyover-game",
    version="0.1",
    description="Shoot as many enemy planes as you can before they get you!",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/stevetasticsteve/Flyover",
    author="Stephen Stanley",
    author_email="stevetasticsteve@gmail.com",
    license="GPL version 3",
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=["pygame"],
    entry_points={
        "console_scripts": [
            "flyover=flyover.__main__:main",
        ]
    },
)



import sys
import random
import os
from importlib import resources

import pygame

import flyover.MapGenerator as MapGenerator
import flyover.Entities as Entities


class GameEngine:
    def __init__(self, settings):
        self.settings = settings
        os.environ['SDL_VIDEO_WINDOW_POS'] = "%d,%d" % (self.settings['window_position'][0],
                                                        self.settings['window_position'][1])
        pygame.init()
        pygame.font.init()
        pygame.mixer.init()

        self.enemy_kill_sfx = []
        for sound in ["explosion_1.ogg", "explosion_2.ogg", "explosion_3.ogg"]:
            with resources.path("flyover.assets.sound_effects.ogg", sound) as path:
                self.enemy_kill_sfx.append(pygame.mixer.Sound(path))

        self.shooting_sfx = []
        for sound in ["shooting_1.ogg", "shooting_2.ogg", "shooting_3.ogg"]:
            with resources.path("flyover.assets.sound_effects.ogg", sound) as path:
                self.shooting_sfx.append(pygame.mixer.Sound(path))

        with resources.path("flyover.assets.sound_effects.ogg", "player_death.ogg") as path:
            self.death_sfx = pygame.mixer.Sound(path)

        self.FPS_clock = pygame.time.Clock()
        self.window = pygame.display.set_mode(self.settings['window_size'])
        pygame.display.set_caption('Flyover')
        pygame.key.set_repeat(10)  # Enables direction button to be held
        self.start_game()

    def start_game(self):
        self.loading_screen()
        self.map = MapGenerator.Map(self.settings['map_size'])
        self.Player = Entities.Player((int(self.settings['window_size'][0] / 2),
                                       int(self.settings['window_size'][1] / 2),
                                       180),
                                      self.settings['map_size'])
        self.entities = self.create_entities()
        self.projectiles = []
        self.Camera = Entities.Camera(self.Player, self.settings['window_size'],
                                      self.map)
        self.score = 0
        self.game_loop()

    def game_loop(self):
        game_over = False
        while True:
            if game_over:
                self.game_over_screen()
            if self.score == self.settings['number_of_enemies'] * 10:
                self.victory_screen()
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    self.close_program()

                self.Player.state = 'level'

                # Key bindings
                # todo can't press two keys at once, like turn and fire at the same time
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_UP:
                        self.Player.accelerate(1)
                    if event.key == pygame.K_DOWN:
                        self.Player.accelerate(-1)
                    if event.key == pygame.K_LEFT:
                        self.Player.rotate(-1)
                        self.Player.state = 'left'
                    if event.key == pygame.K_RIGHT:
                        self.Player.rotate(1)
                        self.Player.state = 'right'
                    if event.key == pygame.K_ESCAPE:
                        self.close_program()
                    if event.key == pygame.K_SPACE:
                        self.projectiles.append(self.Player.shoot_missile())
                        random.choice(self.shooting_sfx).play()

            # Player update
            self.Player.move()
            self.reset_screen()
            if self.Player.state == 'right':
                sprite = self.Player.sprite_right
            elif self.Player.state == 'left':
                sprite = self.Player.sprite_left
            else:
                sprite = self.Player.sprite_level

            player_sprite = pygame.transform.rotate(sprite, self.Player.angle * -1)
            self.window.blit(player_sprite, (int(self.settings['window_size'][0] / 2 - 32),
                                             int(self.settings['window_size'][1] / 2 - 32)))

            # Enemy updates
            for entity in self.entities:
                if entity.last_behaviour_time > self.settings['FPS'] * self.settings['enemy_behaviour_time']:
                    entity.choose_behaviour(self.Player)
                    entity.last_behaviour_time = 0
                else:
                    entity.behaviour(player=self.Player)
                    entity.move()
                shoot = entity.check_sights(self.Player)
                if shoot:
                    self.projectiles.append(shoot)
                    random.choice(self.shooting_sfx).play()

                # Draw enemies
                if entity.within_active_area(self.Camera):
                    if entity.state == 'right':
                        enemy_sprite = pygame.transform.rotate(entity.sprite_right, entity.angle * -1)
                    elif entity.state == 'left':
                        enemy_sprite = pygame.transform.rotate(entity.sprite_left, entity.angle * -1)
                    else:
                        enemy_sprite = pygame.transform.rotate(entity.sprite_level, entity.angle * -1)

                    self.window.blit(enemy_sprite, (self.convert_to_screen_coordinates
                                                    ((entity.blit_x, entity.blit_y))))

            # Projectile updates
            for projectile in self.projectiles:
                if projectile.time_alive > projectile.fuse:
                    del projectile
                    continue
                projectile.move()
                # Check player hits
                if projectile.check_hits(self.Player):
                    if projectile.shooter != str(self.Player):
                        self.death_sfx.play()
                        if not self.settings['invulnerable']:
                            game_over = True
                # Check enemy hits
                for entity in self.entities:
                    if projectile.check_hits(entity):
                        if projectile.shooter == str(self.Player):  # only allow player to hit enemies
                            entity.hit = True
                            random.choice(self.enemy_kill_sfx).play()
                            self.score += 10
                if projectile.within_active_area(self.Camera):
                    self.window.blit(projectile.surface,
                                     (self.convert_to_screen_coordinates((projectile.x, projectile.y))))

                    # delete hit enemies
                    self.entities[:] = [entity for entity in self.entities if not entity.hit]

            # UI
            if self.settings['debug_mode']:
                self.debug_window('Player x = ' + str(self.Player.x) + '   '
                                                                       'Player y = ' + str(self.Player.y))
            self.score_overlay()

            # Game update
            pygame.display.update()
            self.FPS_clock.tick(self.settings['FPS'])

    def loading_screen(self):
        window_size = self.settings['window_size']
        window_x = window_size[0]
        window_y = window_size[1]
        background = pygame.Surface(window_size)
        background.fill((0, 0, 0))
        game_over_font = pygame.font.SysFont(self.settings['game_font'], size=30)

        text_surf = game_over_font.render('Loading', False, (255, 255, 255))
        background.blit(text_surf, (window_x / 2 - 100, window_y / 2))
        self.window.blit(background, (0, 0))
        pygame.display.update()

    def victory_screen(self):
        while True:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    self.close_program()

                # Key bindings
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_r:
                        self.start_game()
                    if event.key == pygame.K_ESCAPE:
                        self.close_program()

            window_size = self.settings['window_size']
            window_x = window_size[0]
            window_y = window_size[1]
            background = pygame.Surface(window_size)
            background.fill((0, 0, 0))
            game_over_font = pygame.font.SysFont(self.settings['game_font'], size=30)

            text_surf = game_over_font.render('You shot down all the planes, you win!', False, (255, 255, 255))
            background.blit(text_surf, (window_x / 2 - 100, window_y / 2))

            text_surf = game_over_font.render('Press r to restart or escape to exit!', False, (255, 255, 255))
            background.blit(text_surf, (window_x / 2 - 150, window_y / 2 + 50))

            text_surf = game_over_font.render('Your score was ' + str(self.score), False, (255, 255, 255))
            background.blit(text_surf, (window_x / 2 - 150, window_y / 2 + 75))

            self.window.blit(background, (0, 0))
            pygame.display.update()
            self.FPS_clock.tick(self.settings['FPS'])

    def game_over_screen(self):
        while True:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    self.close_program()

                # Key bindings
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_r:
                        self.start_game()
                    if event.key == pygame.K_ESCAPE:
                        self.close_program()

            window_size = self.settings['window_size']
            window_x = window_size[0]
            window_y = window_size[1]
            background = pygame.Surface(window_size)
            background.fill((0, 0, 0))
            game_over_font = pygame.font.SysFont(self.settings['game_font'], size=30)

            text_surf = game_over_font.render('Game over!', False, (255, 255, 255))
            background.blit(text_surf, (window_x / 2 - 100, window_y / 2))

            text_surf = game_over_font.render('Press r to restart or escape to exit!', False, (255, 255, 255))
            background.blit(text_surf, (window_x / 2 - 150, window_y / 2 + 50))

            text_surf = game_over_font.render('Your score was ' + str(self.score), False, (255, 255, 255))
            background.blit(text_surf, (window_x / 2 - 150, window_y / 2 + 75))

            self.window.blit(background, (0, 0))
            pygame.display.update()
            self.FPS_clock.tick(self.settings['FPS'])

    def close_program(self):
        pygame.quit()
        sys.exit()

    def reset_screen(self):
        self.window.fill((0, 0, 0))
        self.Camera.move()
        background = self.Camera.active_tiles()
        for tile in background:
            # Need to turn game coordinates into screen coordinates relative to the player
            screen_coord = self.convert_to_screen_coordinates((tile[0][0], tile[0][1]))
            self.window.blit(tile[1], screen_coord)

    def create_entities(self):
        entities = []
        for i in range(self.settings['number_of_enemies']):
            x = random.randint(0, self.settings['map_size'][0])
            y = random.randint(0, self.settings['map_size'][1])
            angle = random.randint(0, 359)
            aircraft = random.choice(['Mig21', 'Mig35'])
            if aircraft == 'Mig21':
                entities.append(Entities.Mig21((x, y, angle), self.settings['map_size']))
            elif aircraft == 'Mig35':
                entities.append(Entities.Mig35((x, y, angle), self.settings['map_size']))
        return entities

    def convert_to_screen_coordinates(self, coord):
        # takes game coordinates and converts it to screen (pixel) coordinates for blitting
        screen_x = coord[0] - self.Camera.rect[0]
        screen_y = coord[1] - self.Camera.rect[1]
        return screen_x, screen_y

    def debug_window(self, content):
        box_size = (200, 100)
        debug_box = pygame.Surface(box_size, flags=pygame.SRCALPHA)
        debug_box.fill((255, 255, 255, 150))
        debug_font = pygame.font.SysFont(self.settings['game_font'], size=18)
        text_surf = debug_font.render(content, False, (0, 0, 0))
        debug_box.blit(text_surf, (10, 10))
        self.window.blit(debug_box, (self.settings['window_size'][0] - box_size[0],
                                     self.settings['window_size'][1] - box_size[1]))

    def score_overlay(self):
        score_bar = pygame.Surface((self.settings['window_size'][0], 25))
        score_bar.fill((255, 255, 255))
        score_font = pygame.font.SysFont(self.settings['game_font'], size=18)
        score_text = score_font.render('Current score: ' + str(self.score), False, (0, 0, 0))
        score_bar.blit(score_text, (10, 2))
        self.window.blit(score_bar, (0, 0))

def main():
    settings = {'FPS': 30,
                'window_size': (1200, 600),
                'window_position': (50, 50),
                'map_size': (6500, 4000),
                'player_start': (100, 100),
                'debug_mode': False,
                'enemy_behaviour_time': 1.5,
                'number_of_enemies': 50,
                'game_font': 'Arial',
                'invulnerable': False}
    game = GameEngine(settings)


if __name__ == '__main__':
    main()